#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Jan. 2008 by Fredo6

# This software is provided as an example of using the Ruby interface to SketchUp.

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name		:   JoinPushPull.rb
# Type		:   Sketchup Tool
# Description	:   push-pull of multiple faces but keeping all generated faces joined
#                                 Vector and Classic Push Pull also included in multi-face
# Menu Items	:   Tools --> "Joint Push Pull" and "Vector Push Pull"
# Context Menu	:   none
# Usage		:   See Tutorial in PDF format
# Original Date	:   20 Jan 2008
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************

require 'sketchup.rb'
require 'LibTraductor.rb'			# for language translation
require 'LibPlaneSelectorTool.rb'	# for custom plane selection
require 'LibVectorSelectorTool.rb'	# for custom vector selection


module JointPushPull

STR_MsgPB_Completion = ["Faces:",
                        "|FR| Faces:"]

STR_MsgNoSelection = ["You must select at least one face",
                      "|FR| Aucune face selectionn\e"]

STR_MsgInvalidSelection = ["Invalid selection for requested Push Pull",
                           "|FR| Selection invalide pour le Push pull demand\"]

STR_MsgDistanceZero = ["Distance is zero",
                       "|FR| La distance est \gale a z\ro"]

DLG_Title_J = ["Joint Push Pull",
			   "|FR| Push-Pull jointif"]
			 
DLG_Title_V = ["Vector Push Pull",
			   "|FR| Push-Pull selon un vecteur"]

DLG_EnumYesNo = { 'Y' => "Yes |FR| Oui", 
				  'N' => "No |FR| Non" }			 

DLG_EnumPlane = { 
                  'N' => ["N: No plane - Direction = normal to face",
                          "|FR| N : Aucun plan - Direction normale \ la face"], 
                  'B' => ["B: Horizontal plane Normal to Blue Axis",
                          "|FR| B : Plan vertical normal \ l'Axe Bleu"], 
				  'R' => ["R: Vertical plane Normal to Red Axis",
          				  "|FR| R : Plan vertical normal \ l'Axe Rouge"],			 
				  'G' => ["G: Vertical plane Normal to Green Axis",
          				  "|FR| G : Plan vertical normal \ l'Axe Vert invers\"],			 
				  'P' => ["P: Custom Plan (last input)",
          				  "|FR| P : Custom Plan (saisie pr\c\dente)"],
				  'C' => ["C: Custom Plan (new input)",
          				  "|FR| C : Custom Plan (nouvelle saisie)"]
				}		  

DLG_EnumVector = { 
                  'N' => ["N: Normal to face (Classic Push Pull)",
                          "|FR| N : Normal \ la face (PushPull classique)"], 
                  'B' => ["B: Blue Axis (Z)",
                          "|FR| B : Axe Bleu (Z)"], 
				  'R' => ["R: Red Axis (X)",
          				  "|FR| R : Axe Rouge (X)"],			 
				  'G' => ["G: Reverse Green Axis (-Y)",
          				  "|FR| G : Axe Vert invers\ (-Y)"],			 
				  'P' => ["P: Custom vector (last input)",
          				  "|FR| P : Vecteur personnalis\e (saisie pr\c\dente)"],
				  'C' => ["C: Custom vector (new input)",
          				  "|FR| C : Vecteur personnalis\e (nouvelle saisie)"]
				 }		  

DLG_EnumFinishing = { 
                      'K' => ["Keep original faces",
                              "|FR| Conserver les faces d'origine"], 
                      'R' => ["Reverse faces (for thickening)",
                              "|FR| Inverser les faces (\paississement)"], 
				      'D' => ["Erase original faces",
          				      "|FR| Effacer les faces d'origine"],			 
     				}		  

DLG_Distance = ["Distance",
                "|FR| Distance"]

DLG_Plane = ["Privileged plane",
             "|FR| Plan privilegi\"]

DLG_Vector = ["Direction Vector",
              "|FR| Vecteur de direction"]
			 
DLG_KeepAsGroup = ["Generate as a Group",
                   "|FR| G\n\rer dans un groupe"]

DLG_Finishing = ["Finishing options",
                 "|FR| Finalisation"]
				   
DLG_Extended = ["Extent influence to non-selected neighbors",
                "|FR| Etendre influence aux voisins non selectionn\s"]

DLG_Borders = ["Create border faces",
               "|FR| Cr\er les faces de bordure"]
				   
DLG_AngleInfluence = ["Angle of influence (degree)",
                      "|FR| Angle d'influence (degr\)"]
				   
NULT_TOOLBAR = "NUL Transformation"
JPP_DIR = "JPP_Dir"
				   
#--------------------------------------------------------------------------------------------------------------
# Top Calling function: create each class once, and reuse them for subsequent calls
#--------------------------------------------------------------------------------------------------------------			 				   
def JointPushPull.execute(type)
	if type == 'V'
		@jppclass_v = JPP.new 'V' unless @jppclass_v
		@jppclass_v.execute
	else
		@jppclass_j = JPP.new 'J' unless @jppclass_j
		@jppclass_j.execute
	end	
end

#--------------------------------------------------------------------------------------------------------------
# Class JointPushPull: implements both Joint and Vector Push Pull
#--------------------------------------------------------------------------------------------------------------			 
class JPP

FaceData = Struct.new("FaceData", :face, :newnormal, :newfaces, :lvd, :embedded, :innerto) 
VxData = Struct.new("VxData", :vertex, :lfaces, :vec, :pt) 

def initialize(type)
	@pp_type = type
	@pp_title = (type == 'V') ? Traductor[DLG_Title_V] : Traductor[DLG_Title_J]
	@tw = Sketchup.create_texture_writer
	@custom_direction_prev = nil

end

def execute
	@model = Sketchup.active_model
    @entities = @model.active_entities
    @selection = @model.selection
	
	#Counting the faces
	n = 0
	@selection.each {|e| n += 1 if e.class == Sketchup::Face}
	if (n == 0)
		UI.beep
		Traductor.messagebox STR_MsgNoSelection
		return
	end	
	
	#Dialog box for parameters
	return unless dialog_parameters
	
	#Processing with calculation
	process_push_pull
end
	
def process_push_pull	
	#building the data structures for faces and vertices
	@hsh_faces = {}
	@hsh_vertices = {}
	@hsh_border_edges = {}
	@hsh_alone_edges = {}
	@embedmax = 0
	@selection.each {|e| compute_data(e) if e.class == Sketchup::Face}
	if (@hsh_faces.length == 0)
		UI.beep
		Traductor.messagebox STR_MsgInvalidSelection
		return
	end		
	
	#Initializing the progress bar
	@pbar = ProgressionBar.new @hsh_faces.length, STR_MsgPB_Completion
	
	#Performing the Joint Push Pull on selected faces
	@model.start_operation @pp_title
		@grp = @entities.add_group 
		@hsh_vertices.each {|item, vd| compute_vertex_offset vd } if (@pp_type == 'J')
		build_all_faces
		@hsh_alone_edges.each { |key, e| @entities.erase_entities e if (e.valid?)} if @option_finishing == 'D'
		@grp.explode unless @option_group
	@model.commit_operation	
end

def dialog_parameters
	unless @dlg
		@hsh_params = {}
		@dlg = Traductor::DialogBox.new @pp_title 
		@dlg.field_numeric "Distance", DLG_Distance, 20.cm, nil, nil
		if (@pp_type == 'J')
			@dlg.field_enum "Direction", DLG_Plane, 'N', DLG_EnumPlane, ['N', 'B', 'R', 'G', 'P', 'C']
		else	
			@dlg.field_enum "Direction", DLG_Vector, 'N', DLG_EnumVector, ['N', 'B', 'R', 'G', 'P', 'C']
		end	
		@dlg.field_enum "Finishing", DLG_Finishing, 'K', DLG_EnumFinishing, ['D', 'R', 'K']
		@dlg.field_enum "Borders", DLG_Borders, 'Y', DLG_EnumYesNo
		@dlg.field_enum "Group", DLG_KeepAsGroup, 'Y', DLG_EnumYesNo
		@dlg.field_enum "Extended", DLG_Extended, 'Y', DLG_EnumYesNo if @pp_type == 'J'
		@dlg.field_numeric "AngleInfluence", DLG_AngleInfluence, 30, 0, 60  if @pp_type == 'J'
	end	

	#Invoking the dialog box
	while true
		return false unless @dlg.show! @hsh_params
		
		@distance = @hsh_params["Distance"]
		break unless (@distance == 0.cm)
		UI.beep
		Traductor.messagebox STR_MsgDistanceZero
	end	

	
	@option_border = (@hsh_params["Borders"] == 'Y') ? true : false
	@option_group = (@hsh_params["Group"] == 'Y') ? true : false
	@option_finishing = @hsh_params["Finishing"]
	if @pp_type == 'J'
		@option_extended = (@hsh_params["Extended"] == 'Y') ? true : false
		@option_angle = @hsh_params["AngleInfluence"].degrees
	end	
	
	@direction = @hsh_params["Direction"]
	case @direction
	when 'N'
		@param_direction = nil
		return true
	when 'B'
		@param_direction = Z_AXIS
		return true
	when 'R'	
		@param_direction = X_AXIS
		return true
	when 'G'	
		@param_direction = Geom::Vector3d.new 0, -1, 0	#reverse Y axis
		return true
	when 'P'	
		@param_direction = @custom_direction_prev	
	when 'C'	
		@param_direction = nil	
	end
	
	#Previous direction
	return true if @param_direction
	
	#Graphical input is required
	if @pp_type == 'J'	
		@model.select_tool LibPlaneSelectorTool::PlaneSelectorTool.new(self, nil, @pp_title)
	else
		@model.select_tool LibVectorSelectorTool::VectorSelectorTool.new(self, @distance, nil, @pp_title)
	end
	@hsh_params["Direction"] = 'P'
	
	false
end

#callback for Plane Selection
def plane_selector_execute(origin, normal)
	@param_direction = normal
	@custom_direction_prev = normal.clone
	process_push_pull
end

#callback for Vector Selection
def vector_selector_execute(vector, distance)
	@param_direction = vector
	@custom_direction_prev = vector.clone
	if distance
		@distance = distance 
		@hsh_params["Distance"] = distance
	end	
	process_push_pull
end

#Build all faces by starting with less embedded faces
def build_all_faces
	@hsh_faces.each { |face_id, fd| optimize_face_with_hole fd }	
	@hsh_faces.each { |face_id, fd| check_inner_face fd }	
	@hsh_faces.each { |face_id, fd| equalize_inner_levels fd }	#to converge the embedded status
	for i in 0..@embedmax
		@hsh_faces.each { |key, fd| build_face fd if fd.embedded == i }
	end	
end

#build data structures for faces and vertex
def compute_data(face)
	#Computing the face data
	newnormal = new_normal face
	return unless newnormal
	@hsh_faces[face.to_s] = fd = FaceData.new
	fd.face = face
	fd.lvd = []
	fd.newnormal = newnormal
	fd.embedded = 0
	fd.innerto = nil
	
	#computing the vertex data
	face.outer_loop.vertices.each do |v|
		vd  = @hsh_vertices[v.to_s]
		unless vd
			@hsh_vertices[v.to_s] = vd = VxData.new
			vd.lfaces = []
		end	
		vd.lfaces.push fd
		vd.vertex = v
		fd.lvd.push vd
	end
end

#check which faces are embedded and add the holes (inner loop) to the data structures
def check_inner_face(fd)
	face = fd.face
	face.loops.each do |l|
		next if l.outer?
		l.vertices.each do |v|
			vd = @hsh_vertices[v.to_s]
			unless vd
				@hsh_vertices[v.to_s] = vd = VxData.new
				vd.vertex = v
			end
			v.faces.each do |f|
				fdi = @hsh_faces[f.to_s]
				fdi.innerto = fd if (fdi && f != face)
			end
		end	
	end
end

def equalize_inner_levels(fd)
	fdi = fd
	n = 0
	while (fdi = fdi.innerto)
		n += 1
	end
	fd.embedded = n
	@embedmax = n unless @embedmax > n
end

#Compute the normal vector to the face, corrected from the privileged plan or vector
def	new_normal(face)
	if @param_direction
		if (@pp_type == 'V')
			newnormal = @param_direction
		else
			pt0 = ORIGIN.offset face.normal, @distance.abs
			newnormal = ORIGIN.vector_to pt0.project_to_plane([ORIGIN, @param_direction])
		end	
	else
		newnormal = face.normal
	end	
	return (newnormal.valid? && ! newnormal.perpendicular?(face.normal)) ? newnormal.normalize : nil
end

#Compute the normal and offset to be used at each vertex in outer loops
def compute_vertex_offset(vd)
	anglelimit = @option_angle
	lvec = []
	vd.lfaces.each {|fd| lvec.push fd.newnormal }
	
	for i in 0..1
		finalvec = []
		
		#Computing the avreage of vectors by groups which are close
		while (lvec.length > 0)
			v0 = lvec[0]
			groupvec = []
			restvec = []
			lvec.each { |v| (v0.angle_between(v) < anglelimit) ? groupvec.push(v) : restvec.push(v) }
			if groupvec.length > 0
				finalvec.push average_vectors(groupvec)
				lvec = restvec
			else
				finalvec = restvec
				break
			end			
		end
		vd.vec = average_vectors finalvec if finalvec.length > 0
		break if i == 1
		
		#adding other faces, as an option, if their normal is not too 'far' from the average
		if @option_extended
			lvec = []
			vd.lfaces.each {|fd| lvec.push fd.newnormal }
			vd.vertex.faces.each do |f| 
				unless (@hsh_faces[f.to_s])
					v = new_normal f
					lvec.push(v) if (v && v.angle_between(vd.vec) < anglelimit)
				end
			end	
		else
			break
		end
	end	

	#computing the length of offset vector
	cosinus = Math::cos vd.lfaces[0].newnormal.angle_between(vd.vec)
	d = (cosinus == 0) ? 1.0 : 1.0 / cosinus
	vd.pt = vd.vertex.position.offset vd.vec, d * @distance
end

#Compute the average of a list of vectors - Return a normalized vector
def average_vectors(lvec)
	x = y = z = 0.0
	lvec.each do |v|
		x += v.x
		y += v.y
		z += v.z
	end
	vec = Geom::Vector3d.new x, y, z
	vec.length = 1.0
	vec
end

#Force planar configuration of faces with holes
def optimize_face_with_hole(fd)
	return if fd.face.loops.length == 1
	pts = []
	fd.lvd.each { |vd| pts.push vd.pt }
	plane = Geom.fit_plane_to_points pts
	fd.lvd.each { |vd| vd.pt = vd.pt.project_to_plane plane}
end

#Build one full face, with borders and holes
def build_face(fd)
	face = fd.face
	face.reverse! if (@option_finishing == 'R' && @distance > 0)
	pts = []
	pts_ref = []
	if (@pp_type == 'V')
		vec = (@direction == 'N') ? face.normal : @param_direction	#Classic and Vector Push pull
		face.outer_loop.vertices.each { |v| @hsh_vertices[v.to_s].pt = v.position.offset vec, @distance}
	end	
	face.outer_loop.vertices.each { |v| pts.push @hsh_vertices[v.to_s].pt }
	face.outer_loop.vertices.each { |v| pts_ref.push v.position }
	if (@option_finishing == 'R')
		pts.reverse!
		pts_ref.reverse!
	end	
	triangulate_face fd, pts, pts_ref
	
	#Building the borders for external faces
	if (@option_border)
		face.outer_loop.edges.each do |e| 
			next if @hsh_border_edges[e.to_s]
			@hsh_border_edges[e.to_s] = e
			n = 0
			e.faces.each { |f| n += 1 if @hsh_faces[f.to_s] }
			if (n == 1) 
				build_edge_border(fd, e) if edge_alone? e
			else
				@hsh_alone_edges[e.to_s] = e
			end	
		end
	end	
	
	#Handling faces with holes
	create_holes fd
	
	#Erasing the original face
	@entities.erase_entities face if (@option_finishing == 'D')
	
	@pbar.countage		#Face treated
end

#Building the new offset face - This may require to triangulate it
def triangulate_face(fd, pts, pts_ref)
	fd.newfaces = []
	lvxnum = []
	n = pts.length - 1
	lvxnum = (0..n).to_a
	polygon_divide(fd.face, fd.newfaces, pts, pts_ref, lvxnum)
end

#Recursive functions to triangulate the face
def polygon_divide(originalface, lstfaces, pts, pts_ref, lvxnum)
	newface = util_make_face originalface, lstfaces, pts, lvxnum
	return if newface
	parts = best_diagonal(pts_ref, lvxnum)
	parts.each { |lvx| polygon_divide originalface, lstfaces, pts, pts_ref, lvx }
end

#generate a portion of the new generated face
def util_make_face(originalface, lstfaces, pts, lvxnum)
	lpts = []
	lvxnum.each { |i| lpts.push pts[i] }
	begin
		newface = @grp.entities.add_face lpts
	rescue
		return nil
	end
	lstfaces.push newface
	transfer_drawing_element originalface, newface
	newface.back_material = originalface.back_material
	transfer_texture originalface, newface, true, lvxnum
	transfer_texture originalface, newface, false, lvxnum
	
	#Transfering the properties of edges
	n = lvxnum.length - 1
	for i in 0..n
		k1 = lvxnum[i]
		k2 = (i == n) ? lvxnum[0] : lvxnum[i+1]
		newedge = newface.outer_loop.edges[i]
		a = (k2 - k1).abs
		if ((a == 1) || (a == pts.length-1))	#Edge belongs to original face
			oldedge = originalface.outer_loop.edges[k1]
			transfer_edge oldedge, newedge
		else									#Edge created for triangulation
			#newedge.visible = false
			newedge.soft = newedge.smooth = true
		end	
	end
	newface
end

#Compute the next best diagonal - This is done on the original face, since it is 'flat'
def best_diagonal(pts_ref, lvxnum)
	nv = lvxnum.length
	nvmax = nv - 3
	pts = []
	lvxnum.each { |i| pts.push pts_ref[i] }
	diag = diffmax = lengmax = nil
	tolerance = 1.1
	
	for i in 0..nvmax
		lim = (i == 0) ? nv - 2 : nv - 1
		for j in (i+2)..lim
			next unless diagonal_valid? pts, i, j
			diffarea = (calculate_area(pts[i..j]) - calculate_area(pts[j..nv-1] + pts[0..i])).abs
			next if (diffmax) && (diffarea >= tolerance * diffmax)
			leng = pts[i].distance pts[j]
			next if (lengmax) && (leng >= tolerance * lengmax)
			diffmax = diffarea
			lengmax = leng
			diag = [i, j]
		end
	end	
	diag = [0, 2] unless diag
	return (diag) ? [lvxnum[diag[0]..diag[1]], lvxnum[diag[1]..nv-1] + lvxnum[0..diag[0]]] : []
end

#Test if a diagonal is valid, that is, in the polygon and not crossing borders
def diagonal_valid?(pts, ibeg, iend)
	#test if middle and points close to diagonal ends are in polygon - Also eliminate colinear diagonals
	[0.1, 0.5, 0.9].each do |v|
		pt = Geom.linear_combination(v, pts[ibeg], 1 - v, pts[iend]) 
		return false unless Geom.point_in_polygon_2D pt, pts, false
	end	
	
	#check if diagonal would cross any edge of the polygon
	n = pts.length-1
	for i in 0..n
		j = (i == n) ? 0 : i+1
		next if ((i == ibeg) || (i == iend) || (j == ibeg) || (j == iend))
		pt = Geom.intersect_line_line [pts[i], pts[j]], [pts[ibeg], pts[iend]]
		return false if ((pt != nil) && ((pt.vector_to pts[ibeg]) % (pt.vector_to pts[iend]) < 0) && 
		                 ((pt.vector_to pts[i]) % (pt.vector_to pts[j]) < 0))
	end	
	true
end

#Calculate the area of a polygone defined by <pts> list of points. This works with concave polygons
def calculate_area(pts)
	n = pts.length - 1
	ptsx = pts + [pts[0]]
	area = 0.0
	for i in 0..n
		j = i+1
		area += ptsx[i].x * ptsx[j].y - ptsx[i].y * ptsx[j].x
	end
	0.5 * area.abs	
end

def vertex_convex? (oldface, kprev, kmid, knext)
	lv = oldface.outer_loop.vertices
	vec1 = lv[kprev].position.vector_to lv[kmid].position
	vec2 = lv[kmid].position.vector_to lv[knext].position
	v = vec1.cross vec2
	return (v.valid? && v % oldface.normal > 0)
end

#Position the texture identical from the original face to the new generated face
def transfer_texture(oldface, newface, front, lvxnum)
	m = (front) ? oldface.material : oldface.back_material
	return unless m && m.texture
	uvh = oldface.get_UVHelper front, !front, @tw
	nv = lvxnum.length - 1
	ptuv = []
	for i in 0..nv
		oldpt = oldface.outer_loop.vertices[lvxnum[i]].position
		kprev = (i ==0) ? lvxnum[nv] : lvxnum[i-1]
		kmid = lvxnum[i]
		knext = (i == nv) ? lvxnum[0] : lvxnum[i+1]
		next unless (vertex_convex? oldface, kprev, kmid, knext)	#Ignore when vertex not convex
		ptuv.push newface.outer_loop.vertices[i].position
		ptuv.push (front) ? uvh.get_front_UVQ(oldpt) : uvh.get_back_UVQ(oldpt)
		break if (ptuv.length == 8)
	end
	begin
		newface.position_material m, ptuv, front
	rescue		#apparently some bugs in Sketchup with projected textures!
	end
end

#just transfer the property of an edge
def transfer_edge(edge, newedge)
	newedge.smooth = edge.smooth?
	newedge.soft = edge.soft?
	transfer_drawing_element edge, newedge
end

#generic transfer for any drawing element
def transfer_drawing_element (old_entity, new_entity)
	new_entity.layer = old_entity.layer
	new_entity.material = old_entity.material
	new_entity.visible = old_entity.visible?
	new_entity.receives_shadows = old_entity.receives_shadows?
	new_entity.casts_shadows = old_entity.casts_shadows?
end

def vertex_alone? (v)
	n = 0
	v.faces.each { |f| n += 1 if @hsh_faces[f.to_s] }
	return (n == 1)
end

def edge_alone? (e)
	n = 0
	e.faces.each { |f| n += 1 if @hsh_faces[f.to_s] }
	return (n == 1)
end

#Build the border of a face edge, either as a 4-vertices face or 2 triangles
def build_edge_border(fd, edge)
	#return unless edge_alone? edge
	face = fd.face
	v1 = edge.start
	v2 = edge.end
	pt1 = v1.position
	pt2 = v2.position
	if (@pp_type == 'V')
		newpt1 = pt1.offset fd.newnormal, @distance
		newpt2 = pt2.offset fd.newnormal, @distance
	else
		newpt1 = @hsh_vertices[v1.to_s].pt
		newpt2 = @hsh_vertices[v2.to_s].pt
	end	
		
	begin
		newface = @grp.entities.add_face [pt1, pt2, newpt2, newpt1]
		reverse_as_needed face, newface, pt1, pt2, edge
		mapping_texture_border(face, newface, pt1, pt2, newpt1, newpt2, true, nil)
		mapping_texture_border(face, newface, pt1, pt2, newpt1, newpt2, false, nil)
	rescue
		newface1 = @grp.entities.add_face [pt1, pt2, newpt2]
		reverse_as_needed face, newface1, pt1, pt2, edge
		
		newface2 = @grp.entities.add_face [newpt2, newpt1, pt1]
		reverse_as_needed face, newface2, pt1, pt2, edge

		mapping_texture_border(face, newface1, pt1, pt2, newpt1, newpt2, true, newface2)
		mapping_texture_border(face, newface1, pt1, pt2, newpt1, newpt2, false, newface2)
		
		#Taking care of the diagonal
		newedges = @grp.entities.add_edges newpt2, pt1
		e = newedges[0]
		transfer_edge edge, e
		#e.visible = false		
		e.soft = e.smooth = true		
	end
	
	#Taking care of the edge for the new face
	newedges = @grp.entities.add_edges pt1, pt2
	transfer_edge edge, newedges[0]
	newedges = @grp.entities.add_edges newpt1, newpt2
	transfer_edge edge, newedges[0]
	
	#Taking care of the edge for the borders
	newedges = @grp.entities.add_edges pt1, newpt1
	e = newedges[0]
	transfer_edge edge, e
	#e.visible = false if edge.curve
	e.soft = e.smooth = true if edge.curve
	newedges = @grp.entities.add_edges pt2, newpt2
	e = newedges[0]
	transfer_edge edge, e
	#e.visible = false if edge.curve
	e.soft = e.smooth = true if edge.curve
end

#Sketchup logic to orientate face is like secret of pyramis!!
def reverse_as_needed (oldface, newface, pt1, pt2, edge)
	transfer_drawing_element oldface, newface
	newface.back_material = oldface.back_material
	ptmid = Geom.linear_combination(0.5, pt1, 0.5, pt2)
	pt = ptmid.offset newface.normal, 0.1
	pt = pt.project_to_plane oldface.plane
	newface.reverse! if (oldface.classify_point(pt) == 1)
	newface.reverse! if (@distance < 0 && @option_finishing != 'R')
end

#Most complex routine to find the right way to map the texture on the borders
def mapping_texture_border(oldface, newface, pt1, pt2, newpt1, newpt2, front, newface2)
	m = (front) ? oldface.material : oldface.back_material
	return nil unless m && m.texture
	uvh = oldface.get_UVHelper front, !front, @tw
	puv1 = (front) ? uvh.get_front_UVQ(pt1) : uvh.get_back_UVQ(pt1)
	puv2 = (front) ? uvh.get_front_UVQ(pt2) : uvh.get_back_UVQ(pt2)
	ptc = oldface.bounds.center
	puvc = (front) ? uvh.get_front_UVQ(ptc) : uvh.get_back_UVQ(ptc)
	
	u1 = puv1.x
	v1 = puv1.y
	u2 = puv2.x
	v2 = puv2.y	

	w = m.texture.width
	h = m.texture.height	
	d = pt1.distance newpt1
	vec = (Geom::Vector3d.new((u2-u1) * w, (v2-v1) * h, 0)) * Z_AXIS
	vec.length = d
	x = vec.x.abs / w
	y = vec.y.abs / h
	#v = y.ceil
	ushift = (x - x.round).abs
	vshift = (y - y.round).abs
	if (vshift > ushift)
		u = x.round
		v = y.ceil
	else
		u = x.ceil
		v = y.round
	end
	usign = (vec.x <=> 0)
	vsign = (vec.y <=> 0)
	sense = ((pt1.vector_to(pt2) * pt1.vector_to(newpt1)) % newface.normal) <=> 0.0
	oldsense = ((pt1.vector_to(pt2) * pt1.vector_to(ptc)) % oldface.normal) <=> 0.0
	uvsense = ((puv1.vector_to(puv2) * puv1.vector_to(puvc)) % Z_AXIS) <=> 0.0

	fac = -uvsense * oldsense * sense
	udec = usign * u * fac
	vdec = vsign * v * fac
	
	#Painting rectangular face or first triangle
	u3 = u1 + udec
	v3 = v1 + vdec
	u4 = u2 + u3 - u1
	v4 = v2 + v3 - v1
	puv3 = Geom::Point3d.new u3, v3, puv1.z
	puv4 = Geom::Point3d.new u4, v4, puv1.z
	ptuv = [pt1, puv1, pt2, puv2, newpt2, puv4, newpt1, puv3]
	newface.position_material m, ptuv, front
	
	#taking care of second triangle
	if (newface2)
		sense2 = ((pt1.vector_to(pt2) * pt1.vector_to(newpt1)) % newface2.normal) <=> 0
		u3 = u1 + udec * sense * sense2
		v3 = v1 + vdec * sense * sense2
		u4 = u2 + u3 - u1
		v4 = v2 + v3 - v1
		puv3 = Geom::Point3d.new u3, v3, puv1.z
		puv4 = Geom::Point3d.new u4, v4, puv1.z
		ptuv = [pt1, puv1, pt2, puv2, newpt2, puv4, newpt1, puv3]
		newface2.position_material m, ptuv, front		
	end
end

#Manage the creation of holes in the generated faces
def create_holes(fd)
	face = fd.face
	return if face.loops.length == 1		#no hole in face
	
	if (@pp_type == 'V')
		face.loops.each { |l| dig_hole_V(fd, l) unless l.outer? }
	elsif (fd.newfaces.length == 1)
		face.loops.each { |l| dig_flat_hole_J(fd, l) unless l.outer? }
	else
		face.loops.each { |l| dig_complex_hole_J(fd, l) unless l.outer? }
	end	
end

#Simple algorithm, as Vector Push Pull preserves faces
def dig_hole_V(fd, loop)		
	pts = []
	loop.vertices.each { |v| pts.push v.position.offset(fd.newnormal, @distance) }
	newface = @grp.entities.add_face pts
	lstedges = []
	loop.edges.each { |e| lstedges.push e if (edge_alone? e) }
	@grp.entities.erase_entities newface unless (lstedges.length == 0)
	lstedges.each { |e|	build_edge_border fd, e } if @option_border	
end

#Algorithm to compute vertex positions for holes and embedded faces when generated face is planar
def dig_flat_hole_J(fd, loop)
	pts = []
	plane = fd.newfaces[0].plane
	loop.vertices.each do |v|
		vd = @hsh_vertices[v.to_s]
		vd.vec = average_vector_loop fd, v
		pt = Geom.intersect_line_plane [v.position, vd.vec], plane 
		pts.push pt
		vd.pt = pt.clone
	end
	newface = @grp.entities.add_face pts
	lstedges = []
	loop.edges.each { |e| lstedges.push e if (edge_alone? e) }
	@grp.entities.erase_entities newface unless (lstedges.length == 0)
	lstedges.each { |e|	build_edge_border fd, e } if @option_border	
end

#Algorithm to compute vertex positions for holes and embedded faces when generated face is triangulated
def dig_complex_hole_J(fd, loop)

end

#Compute position of vertices for loops, based on embedding face
def average_vector_loop(fd, v)
	pt = v.position
	x = y = z = 0.0
	n = fd.lvd.length - 1
	for i in 0..n
		j = (i == n) ? 0 : i + 1
		vd1 = fd.lvd[i]
		vd2 = fd.lvd[j]
		d1 = pt.distance vd1.vertex.position
		d2 = pt.distance vd2.vertex.position
		d = pt.distance_to_line [vd1.vertex.position, vd2.vertex.position]
		vv = Geom::Vector3d.linear_combination(d2, vd1.vec, d1, vd2.vec)
		fac = 1 / d
		x += vv.x * fac
		y += vv.y * fac
		z += vv.z * fac
	end
	vec = Geom::Vector3d.new x, y, z
	vec.length = 1.0
	vec
end

end #class JPP

#--------------------------------------------------------------------------------------------------------------
# Class ProgressionBar: progress bar in the Sketchup Status text area
#--------------------------------------------------------------------------------------------------------------
class ProgressionBar

#Initialization of progress bar
def initialize(nbelts, label)
	@pb_nbelts = nbelts
	@pb_label = Traductor[label]
	@pb_progression = 0
	@pb_rangemax = 200
	@pb_range = 0
	@pb_time0 = Time.now	
end

#Increment the Progression Bar by <nb> steps
def countage(nb=1)
	@pb_progression += nb
	f = 100 * @pb_progression / @pb_nbelts
	percent = f.to_i
	if (percent != @pb_range)
		@pb_range = percent
		n = percent * @pb_rangemax / 100
		Sketchup::set_status_text "|" * n.to_i
	end	
	Sketchup.set_status_text @pb_label + " #{@pb_progression} / #{@pb_nbelts}", SB_VCB_LABEL
	Sketchup::set_status_text "#{@pb_range}%  -  #{sprintf "%4.2f", Time.now - @pb_time0} sec", SB_VCB_VALUE
end

end #class ProgressionBar

#--------------------------------------------------------------------------------------------------------------
# Public methods of module
#--------------------------------------------------------------------------------------------------------------

def JointPushPull.set_command
	menutool = UI.menu "Tools"
	menutool.add_separator
	@tlb = nil
	
	#Joint Push Pull Command
	cmd = UI::Command.new(Traductor[DLG_Title_J]) { JointPushPull.execute 'J' }
	menutool.add_item cmd
	tooltip = Traductor[DLG_Title_J]
	JointPushPull.create_button cmd, tooltip, "JPP_J.png"
	
	#Vector Push Pull command
	cmd = UI::Command.new(Traductor[DLG_Title_V]) { JointPushPull.execute 'V' }
	menutool.add_item cmd
	tooltip = Traductor[DLG_Title_V]
	JointPushPull.create_button cmd, tooltip, "JPP_V.png"
	
	#showing the toolbar
	@tlb.show if @tlb
end

def JointPushPull.create_button(cmd, tooltip, iconpath)
	cmd.status_bar_text = tooltip
	icon = Sketchup.find_support_file iconpath, "Plugins/" + JPP_DIR
	icon = Sketchup.find_support_file iconpath, "Plugins" unless icon
	if (icon)
		cmd.tooltip = tooltip
		cmd.small_icon = cmd.large_icon = icon
		@tlb = UI::Toolbar.new NULT_TOOLBAR unless @tlb
		@tlb.add_item cmd
		@tlb.show
	end	

end

unless $jpp____loaded
	JointPushPull.set_command
	$jpp____loaded = true
end

end #Module JointPushPull
